from thermostat import Thermostat

import pandas as pd
import numpy as np
from eemeter.location import zipcode_to_station
from eemeter.weather import ISDWeatherSource
from eemeter.evaluation import Period

import warnings
from datetime import datetime
from datetime import timedelta
import dateutil.parser
import os

def from_csv(metadata_filename, verbose=False):
    """
    Creates Thermostat objects from data stored in CSV files.

    Parameters
    ----------
    metadata_filename : str
        Path to a file containing the thermostat metadata.
    verbose : boolean
        Set to True to output a more detailed log of import activity.

    Returns
    -------
    thermostats : iterator over thermostat.Thermostat objects
        Thermostats imported from the given CSV input files.
    """
    metadata = pd.read_csv(metadata_filename,
            dtype={"thermostat_id": str, "zipcode": str, "utc_offset": str, "equipment_type": int, "interval_data_filename": str})


    for i, row in metadata.iterrows():
        if verbose:
            print("Importing thermostat {}".format(row.thermostat_id))

        # make sure this thermostat type is supported.
        if row.equipment_type not in [1,2,3,4,5]:
            warnings.warn("Skipping import of thermostat controlling equipment"
                          " of unsupported type. (id={})".format(row.thermostat_id))
            continue

        interval_data_filename = os.path.join(os.path.dirname(metadata_filename), row.interval_data_filename)

        thermostat = get_single_thermostat(
                row.thermostat_id,
                row.zipcode,
                row.equipment_type,
                row.utc_offset,
                interval_data_filename
        )

        yield thermostat

def get_single_thermostat(thermostat_id, zipcode, equipment_type, utc_offset, interval_data_filename):
    df = pd.read_csv(interval_data_filename)

    heating, cooling, aux_emerg = _get_equipment_type(equipment_type)

    # load indices
    dates = pd.to_datetime(df["date"])
    daily_index = pd.DatetimeIndex(start=dates[0], periods = dates.shape[0], freq="D")
    hourly_index = pd.DatetimeIndex(start=dates[0], periods = dates.shape[0] * 24, freq="H")

    # raise an error if dates are not aligned
    if not all(dates == daily_index):
        message("Dates provided for thermostat_id={} may contain some "
                "which are out of order, missing, or duplicated.".format(thermostat_id))
        raise ValueError(message)

    # load hourly time series values
    temp_in = pd.Series(_get_hourly_block(df, "temp_in"), hourly_index)

    if heating:
        heating_setpoint = pd.Series(_get_hourly_block(df, "heating_setpoint"), hourly_index)
    else:
        heating_setpoint = None

    if cooling:
        cooling_setpoint = pd.Series(_get_hourly_block(df, "cooling_setpoint"), hourly_index)
    else:
        cooling_setpoint = None

    if aux_emerg:
        auxiliary_heat_runtime = pd.Series(_get_hourly_block(df, "auxiliary_heat_runtime"), hourly_index)
        emergency_heat_runtime = pd.Series(_get_hourly_block(df, "emergency_heat_runtime"), hourly_index)
    else:
        auxiliary_heat_runtime = None
        emergency_heat_runtime = None

    # load outdoor temperatures
    station = zipcode_to_station(zipcode)
    ws_hourly = ISDWeatherSource(station, daily_index[0].year, daily_index[-1].year)
    utc_offset = dateutil.parser.parse("2000-01-01T00:00:00" + utc_offset).tzinfo.utcoffset(None)
    temp_out = pd.Series(_get_outdoor_temperatures(daily_index, ws_hourly, utc_offset), hourly_index)

    # load daily time series values
    if cooling:
        cool_runtime = pd.Series(df["cool_runtime"].values, daily_index)
    else:
        cool_runtime = None
    if heating:
        heat_runtime = pd.Series(df["heat_runtime"].values, daily_index)
    else:
        heat_runtime = None

    # create thermostat instance
    thermostat = Thermostat(thermostat_id, equipment_type, zipcode, station,
            temp_in, temp_out, cooling_setpoint, heating_setpoint,
            cool_runtime, heat_runtime, auxiliary_heat_runtime,
            emergency_heat_runtime)
    return thermostat

def _get_hourly_block(df, prefix):
    columns = ["{}_{:02d}".format(prefix, i) for i in range(24)]
    values = df[columns].values
    return values.reshape((values.shape[0] * values.shape[1],))

def _get_outdoor_temperatures(dates, ws_hourly, utc_offset):
    all_temps = []
    for date in dates:
        date = date - utc_offset
        period = Period(date, date + timedelta(days=1))
        temps = ws_hourly.hourly_temperatures(period, "degF")
        all_temps.append(temps)
    return np.array(all_temps).reshape(dates.shape[0] * 24)

def _get_equipment_type(equipment_type):
    """
    Returns
    -------
    heating : boolean
        True if the equipment type has heating equipment
    cooling : boolean
        True if the equipment type has cooling equipment
    aux_emerg : boolean
        True if the equipment type has auxiliary/emergency heat equipment
    """
    if equipment_type == 1:
        return True, True, True
    elif equipment_type == 2:
        return True, True, False
    elif equipment_type == 3:
        return True, True, False
    elif equipment_type == 4:
        return True, False, False
    elif equipment_type == 5:
        return False, True, False
    else:
        return None



import numpy as np

def get_daily_avoided_runtime(alpha, demand, demand_baseline):
    """ Calculate avoided runtime from an estimate of alpha, demand and demand
    baseline for heating or cooling seasons. Expects the sign of demand to be
    such that it correlates positively with runtime.

    Parameters
    ----------
    alpha : float
        Estimate of slope of runtime response to demand.
    demand : pandas.Series
        Timeseries of daily demand for a particular season.
    demand_baseline : pandas.Series
        Timeseries of baseline daily demand for a particular season.

    Returns
    -------
    daily_avoided_runtime : pandas.Series
        Timeseries of daily avoided runtime.
    """
    return alpha * (demand_baseline - demand)

def get_seasonal_percent_savings(total_baseline_runtime, daily_avoided_runtime):
    """ Estimate baseline heating runtime given a daily avoided runtimes.

    Parameters
    ----------
    total_baseline_runtime : pandas.Series
        Total baseline season runtime for a particular season
    daily_avoided_runtime : pandas.Series
        Timeseries of daily avoided runtime for a particular season.

    Returns
    -------
    seasonal_percent_savings : pandas.Series
        Seasonal percent savings as given by :math:`\\text{seasonal percent
        savings} = \\frac{\\text{seasonal avoided runtime}}{\\text{baseline
        seasonal runtime}}`. Returned as a decimal.
    """
    total_avoided_runtime = np.nansum(daily_avoided_runtime)

    seasonal_percent_savings = total_avoided_runtime / float(total_baseline_runtime)
    return seasonal_percent_savings



import numpy as np
from numpy.linalg import lstsq

def runtime_regression(daily_runtime, daily_demand):
    """
    Least squares regession of runtime against a measure of demand.

    Parameters
    ----------
    hourly_runtime : pd.Series with pd.DatetimeIndex
        Runtimes for a particular heating or cooling season.
    daily_demand : pd.Series with pd.DatetimeIndex
        A daily demand measure for each day in the heating or cooling season.

    Returns
    -------
    slope : float
        The slope parameter found by the regression to minimize sq error
    intercept : float
        The intercept parameter found by the regression to minimize sq error
    mean_sq_err : float
        The mean squared error of the regession.
    """
    if daily_demand.shape[0] == 0 and daily_runtime.shape[0] == 0:
        return np.nan, np.nan

    x = daily_demand.values[:,np.newaxis]
    y = daily_runtime
    results = lstsq(x, y) # model: y = a*x
    slope = results[0][0]
    if daily_demand.shape[0] > 1 and daily_runtime.shape[0] > 1:
        mean_sq_err = results[1][0] / y.shape[0] # convert from sum sq err
    else:
        mean_sq_err = np.nan
    return slope, mean_sq_err




import pandas as pd
import numpy as np

from collections import OrderedDict

REAL_OR_INTEGER_VALUED_COLUMNS = [
    "n_days_both_heating_and_cooling",
    "n_days_insufficient_data",
    "n_days_in_season",
    "n_days_in_season_range",
    "slope_deltaT",
    "alpha_est_dailyavgCDD",
    "alpha_est_dailyavgHDD",
    "alpha_est_hourlyavgCDD",
    "alpha_est_hourlyavgHDD",
    "mean_sq_err_dailyavgCDD",
    "mean_sq_err_dailyavgHDD",
    "mean_sq_err_hourlyavgCDD",
    "mean_sq_err_hourlyavgHDD",
    "mean_squared_error_deltaT",
    "deltaT_base_est_dailyavgCDD",
    "deltaT_base_est_dailyavgHDD",
    "deltaT_base_est_hourlyavgCDD",
    "deltaT_base_est_hourlyavgHDD",
    "baseline_daily_runtime_deltaT",
    "baseline_daily_runtime_dailyavgCDD",
    "baseline_daily_runtime_dailyavgHDD",
    "baseline_daily_runtime_hourlyavgCDD",
    "baseline_daily_runtime_hourlyavgHDD",
    "baseline_seasonal_runtime_deltaT",
    "baseline_seasonal_runtime_dailyavgCDD",
    "baseline_seasonal_runtime_dailyavgHDD",
    "baseline_seasonal_runtime_hourlyavgCDD",
    "baseline_seasonal_runtime_hourlyavgHDD",
    "baseline_comfort_temperature",
    "actual_daily_runtime",
    "actual_seasonal_runtime",
    "seasonal_avoided_runtime_deltaT",
    "seasonal_avoided_runtime_dailyavgCDD",
    "seasonal_avoided_runtime_dailyavgHDD",
    "seasonal_avoided_runtime_hourlyavgCDD",
    "seasonal_avoided_runtime_hourlyavgHDD",
    "seasonal_savings_deltaT",
    "seasonal_savings_dailyavgCDD",
    "seasonal_savings_dailyavgHDD",
    "seasonal_savings_hourlyavgCDD",
    "seasonal_savings_hourlyavgHDD",
    "rhu_00F_to_05F",
    "rhu_05F_to_10F",
    "rhu_10F_to_15F",
    "rhu_15F_to_20F",
    "rhu_20F_to_25F",
    "rhu_25F_to_30F",
    "rhu_30F_to_35F",
    "rhu_35F_to_40F",
    "rhu_40F_to_45F",
    "rhu_45F_to_50F",
    "rhu_50F_to_55F",
    "rhu_55F_to_60F",
]

def combine_output_dataframes(dfs):
    """ Combines output dataframes. Useful when combining output from batches.

    Parameters
    ----------
    dfs : list of pd.DataFrame
        Output dataFrames to combine into one.

    Returns
    -------
    out : pd.DataFrame
        Dataframe with combined output metadata.

    """
    return pd.concat(dfs, ignore_index=True)

def compute_summary_statistics(df, label):
    """ Computes summary statistics for the output dataframe. Computes the
    following statistics for each real-valued or integer valued column in
    the output dataframe: mean, standard error of the mean, 10th, 20th, 30th,
    40th, 50th, 60th, 70th, 80th, and 90th percentiles.

    Parameters
    ----------
    df : pd.DataFrame
        Output for which to compute summary statistics.
    label : str
        Name for this set of thermostat outputs.

    Returns
    -------
    stats : collections.OrderedDict
        An ordered dict containing the summary statistics. Column names are as
        follows, in which ### is a placeholder for the name of the column:

          - mean: ###_mean
          - standard error of the mean: ###_sem
          - 10th quantile: ###_10q
          - 20th quantile: ###_20q
          - 30th quantile: ###_30q
          - 40th quantile: ###_40q
          - 50th quantile: ###_50q
          - 60th quantile: ###_60q
          - 70th quantile: ###_70q
          - 80th quantile: ###_80q
          - 90th quantile: ###_90q

    """
    quantiles = [10, 20, 30, 40, 50, 60, 70, 80, 90]

    stats = OrderedDict()
    stats["label"] = label
    for column_name in REAL_OR_INTEGER_VALUED_COLUMNS:
        column = df[column_name]

        # calculate quantiles
        stats["{}_mean".format(column_name)] = np.nanmean(column)
        stats["{}_sem".format(column_name)] = np.nanstd(column) / (column.count() ** .5)
        for quantile in quantiles:
            stats["{}_q{}".format(column_name, quantile)] = column.quantile(quantile / 100.)

    return stats

def summary_statistics_to_csv(stats, filepath):
    """ Write metric statistics to CSV file.

    Parameters
    ----------
    stats : list of dict
        List of outputs from thermostat.stats.compute_summary_statistics()

    """
    quantiles = [10, 20, 30, 40, 50, 60, 70, 80, 90]
    columns = ["label"]
    for column_name in REAL_OR_INTEGER_VALUED_COLUMNS:
        columns.append("{}_mean".format(column_name))
        columns.append("{}_sem".format(column_name))
        for quantile in quantiles:
            columns.append("{}_q{}".format(column_name, quantile))

    stats_dataframe = pd.DataFrame(stats, columns=columns)
    stats_dataframe.to_csv(filepath, index=False, columns=columns)
    return stats_dataframe



from setuptools import setup, find_packages, Command

version = __import__('thermostat').get_version()

long_description = "Calculate connected thermostat temperature/run-time savings."

class PyTest(Command):
    user_options = []
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        import sys
        errno = subprocess.call([sys.executable, 'runtests.py', '--runslow', '--cov-report', 'term-missing', '--cov', 'thermostat'  ])
        raise SystemExit(errno)


setup(name='thermostat',
    version=version,
    description='Calculate connected thermostat savings',
    long_description=long_description,
    url='https://github.com/impactlab/thermostat/',
    author='Phil Ngo',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    cmdclass = {'test': PyTest},
    keywords='thermostat savings EPA',
    packages=find_packages(),
    install_requires=[
        'eemeter',
        'pandas',
        ],
)
